/* UNH CS 758 assignment 3

   array-based dictionary for char * keys and void * values
   
   a slightly-generalized implementation of the dict data structure
     declared in babble.h - no reason not to have it handle generic
     void * instead of the specific dist * declared in babble.h
   
   this implementation is based on an unsorted array, so it will have
     linear time lookup and amortized constant time insertion :(.

   the implementation is similar to that of dist in main.c, since they
     are fundamentall the same data structure
*/

#include <stdlib.h>  /* malloc */
#include <string.h>  /* strcmp */
#include "utils.h"
#include "babble.h"


/******** first, we declare the types that will be used to hold the data ***/


typedef struct {
  char *key;
  void *value;
} array_dict_entry;


typedef struct {
  int size;
  int len;
  /* here's the (resizable) array */
  array_dict_entry *entries;
} array_dict;


/********** now for the functions to manipulate those types *************/
  

/* returns nonzero if key not found, otherwise sets val */
int array_dict_lookup(array_dict *ad, char *key, void **valptr)
{
  int i;
  array_dict_entry *es = ad->entries;

  for (i = 0; i < ad->len; i++) {
    if (strcmp(key, es[i].key) == 0) {
      *valptr = es[i].value;
      return 0;
    }
  }
  return 1;
}


/* returns nonzero if insertion failed (eg out of memory).  does not
   check if key is already present */
int array_dict_insert(array_dict *ad, char *key, void *val)
{
  array_dict_entry *e;

  if (ad->len == ad->size) {
    int newsize = ad->size * 2;
    ad->entries = (array_dict_entry *) realloc(ad->entries, sizeof(array_dict_entry) * newsize);
    if (! ad->entries) return EXIT_FAILURE;
    /* panic("failed to increase size of entries in array_dict"); */
    ad->size = newsize;
  }
  e = &((ad->entries)[ad->len++]);
  e->key = key;
  e->value = val;
  return EXIT_SUCCESS;
}


void array_dict_iter(array_dict *ad, void (*f)(char *, void *))
{
  int i;

  for (i = 0; i < ad->len; i++) {
    array_dict_entry e = (ad->entries)[i];
    f(e.key, e.value);
  }
}


/* for debugging. assumes values are dist *  */
void array_dict_fprint(FILE *outfile, array_dict *ad)
{
  int i;

  fprintf(outfile, "-- array dict of size %d with %d entries\n",
	  ad->size, ad->len);
  for (i = 0; i < ad->len; i++) {
    array_dict_entry e = ad->entries[i];
    fprint_pair(outfile, e.key, (dist *) e.value);
  }
  fprintf(outfile, "-- end of array dict\n");
}


void array_dict_free(array_dict *ad)
{
  free(ad->entries);
  free(ad);
}


dict *make_array_dict(void)
{
  array_dict_entry *es;
  array_dict *ad;
  dict *d;
  
  es = (array_dict_entry *) malloc(sizeof(array_dict_entry) * DICT_INIT_SIZE);
  if (!es) panic("failed to malloc initial array_dict entries");
  ad = (array_dict *) malloc(sizeof(array_dict));
  if (!ad) panic("failed to malloc array_dict");
  ad->size = DICT_INIT_SIZE;
  ad->len = 0;
  ad->entries = es;
  d = (dict *) malloc(sizeof(dict));
  if (!d) panic("failed to malloc initial dict");
  d->data = ad;
  /* our functions handle void *; cast to functions handling dist * */
  d->lookup = (lookup_t) array_dict_lookup;
  d->insert = (insert_t) array_dict_insert;
  d->iter = (iter_t) array_dict_iter;
  d->fprint = (fprint_t) array_dict_fprint;
  d->free = (free_t) array_dict_free;
  return d;
}


/* EOF */
