/* UNH CS 758 assignment 3

   Record trigram probabilities from input and generate text.

   The trigram model is stored as a distribution over the third token
   given the first two tokens.  The first two tokens are concatenated
   to form a "key" in a dictionary/association list/map-style data
   structure, for which the distribution is the associated "value".
   This implementation is written such that the dictionary data
   structure can be implemented in multiple different ways, allowing
   performance to be compared.

   areas for possible improvement:
     rewrite harness: arg order, stdout vs stderr
     implement linked_list
     wrap output, remove spaces before punctuation
     tokenization of contractions
     separate timing of model building and generation

   re-written from scratch by Wheeler Ruml, September 2023
*/


#include <stdio.h>  /* FILE */


/* how many times a token occured */
typedef struct {
  char *token;
  int  count;
  int  to_right; /* used for generation: total counts to the right of
		    this entry */
} count;


/* distribution of counts over tokens
   
   storing the counts as an array means linear time :( we are assuming
   few tokens per key */
typedef struct {
  int size;  /* physical size of counts array */
  int len;   /* how many elements are current stored */
  count *counts;
} dist;

#define DIST_INIT_SIZE 2

  
/* a generic dictionary data structure that maps char * to dist *
   (strings to distributions)

   to allow the same code to use different implementations of the data
   structure, we package the data structure itself (held as a void *)
   with the functions that should be used to manipulate it (more
   precisely, pointers to those functions).  different implementations
   of the data structure can fill in the appropriate functions for
   using it.  (this is how many object-oriented languages are
   implemented.)  if it has been a while since you last saw function
   pointers, you might want to refresh.

   note that the dict doesn't copy the keys and values - it assumes
   that the pointers it is given will remain valid and that the data
   pointed to will be explicitly freed by the user when the dict is no
   longer needed.

   lookup returns nonzero if key not found, otherwise sets val
   insert returns nonzero if insertion failed (eg out of memory). does not
      check if key is already present
   iter calls the given function on all key/value pairs
   fprint is for printing hash bucket occupancies or debugging
   free frees whatever is pointed to by data (but not the keys and values)

   first, for clarity, typedefs for the types of the interface functions */
typedef int (*lookup_t)(void *d, char *key, dist **valptr);   /* returns err */
typedef int (*insert_t)(void *d, char *key, dist *val);    /* returns err */
typedef void (*iter_t)(void *d, void (*f)(char *key, dist *val)); /* takes function pointer */
typedef void (*fprint_t)(FILE *outfile, void *d);
typedef void (*free_t)(void *d);

/* now the dict itself */
typedef struct dict {
  void *data;
  lookup_t lookup;
  insert_t insert;
  iter_t iter;
  fprint_t fprint;
  free_t free;
} dict;

#define DICT_INIT_SIZE 8


/* defined in main.c */
void fprint_pair(FILE*, char *, dist *);


/* EOF */
