/* UNH CS 758 assignment 3

   linked list-based dictionary for char * keys and void * values
   
   linear time lookup and constant time insertion :(.
*/

#include <stdlib.h>  /* malloc */
#include <string.h>  /* strcmp */
#include "utils.h"
#include "babble.h"


/******** first, we declare the types that will be used to hold the data ***/


typedef struct ll_node {
  char *key;
  void *value;
  struct ll_node *next;
} ll_node;


typedef struct ll_dict {
  ll_node *list;
} ll_dict;


/********** now for the functions to manipulate those types *************/
  

int ll_lookup(ll_node *n, char *key, void **valptr)
{
  if (n) {
    if (strcmp(key, n->key) == 0) {
      *valptr = n->value;
      return 0;
    } else
      return ll_lookup(n->next, key, valptr);
  }
  return 1;
}


/* returns nonzero if key not found, otherwise sets val */
int ll_dict_lookup(ll_dict *lld, char *key, void **valptr)
{
  return ll_lookup(lld->list, key, valptr);
}


/* returns nonzero if insertion failed (eg out of memory).  does not
   check if key is already present */
int ll_dict_insert(ll_dict *lld, char *key, void *val)
{
  ll_node *n = (ll_node *) malloc(sizeof(ll_node));

  if (!n) return EXIT_FAILURE;
  n->key = key;
  n->value = val;
  n->next = lld->list;
  lld->list = n;
  return EXIT_SUCCESS;
}


void ll_iter(ll_node *n, void (*f)(char *, void *))
{
  if (n) {
    f(n->key, n->value);
    ll_iter(n->next, f);
  }
}


void ll_dict_iter(ll_dict *lld, void (*f)(char *, void *))
{
  ll_iter(lld->list, f);
}


void ll_fprint(FILE *outfile, ll_node *n)
{
  if (n) {
    fprint_pair(outfile, n->key, (dist *) n->value);
    ll_fprint(outfile, n->next);
  }
}


/* for debugging. assumes values are dist *  */
void ll_dict_fprint(FILE *outfile, ll_dict *lld)
{
  fprintf(outfile, "-- linked list dict\n");
  ll_fprint(outfile, lld->list);
  fprintf(outfile, "-- end of linked list dict\n");
}


void ll_free(ll_node *n)
{
  if (n) {
    ll_node *next = n->next;
    free(n);
    ll_free(next);
  }
}


void ll_dict_free(ll_dict *lld)
{
  ll_free(lld->list);
  free(lld);
}


dict *make_ll_dict(void)
{
  ll_dict *lld;
  dict *d;
  
  lld = (ll_dict *) malloc(sizeof(ll_dict));
  if (!lld) panic("failed to malloc ll_dict");
  lld->list = NULL;
  d = (dict *) malloc(sizeof(dict));
  if (!d) panic("failed to malloc initial dict");
  d->data = lld;
  /* our functions handle void *; cast to functions handling dist * */
  d->lookup = (lookup_t) ll_dict_lookup;
  d->insert = (insert_t) ll_dict_insert;
  d->iter = (iter_t) ll_dict_iter;
  d->fprint = (fprint_t) ll_dict_fprint;
  d->free = (free_t) ll_dict_free;
  return d;
}


/* EOF */
